"""
Data Types for Sutton SignWriting core functionality.
"""

from typing import (
    TypedDict,
    List,
    Union,
    Optional,
    Dict,
    NotRequired,
)

# Query-related types


class QuerySignboxSymbol(TypedDict):
    """A symbol in a signbox query."""

    symbol: str
    """A symbol identifier."""
    coord: Optional[List[int]]
    """An optional coordinate [x, y]."""


class QuerySignboxRange(TypedDict):
    """A symbol range in a signbox query."""

    range: List[str]
    """An array of two symbols defining the range [start, end]."""
    coord: Optional[List[int]]
    """An optional coordinate [x, y]."""


class QuerySignboxOr(TypedDict):
    """An OR group of symbols or ranges in a signbox query."""

    or_: List[Union[str, List[str]]]
    """An array of symbol strings and/or range arrays."""
    coord: Optional[List[int]]
    """An optional coordinate [x, y]."""


QuerySignboxElement = Union[QuerySignboxSymbol, QuerySignboxRange, QuerySignboxOr]


class QueryPrefix(TypedDict):
    """Prefix elements for a query."""

    required: bool
    """Whether the prefix is required."""
    parts: Optional[List[Union[str, List[str], List[Union[str, List[str]]]]]]
    """An object for prefix elements."""


class QueryObject(TypedDict):
    """
    Object of query elements with regular expression identification.
    """

    query: bool
    """Required true for query object."""
    prefix: NotRequired[Optional[QueryPrefix]]
    """An object for prefix elements."""
    signbox: NotRequired[Optional[List[QuerySignboxElement]]]
    """Array of objects for symbols, ranges, and list of symbols or ranges, with optional coordinates."""
    variance: NotRequired[Optional[int]]
    """Amount that x or y coordinates can vary and find a match, defaults to 20."""
    style: NotRequired[Optional[bool]]
    """Boolean value for including style string in matches."""


# Column and layout types


class ColumnPunctuation(TypedDict):
    """Punctuation options for columns."""

    spacing: Optional[bool]
    """Whether to add spacing for punctuation."""
    pad: Optional[int]
    """Padding for punctuation."""
    pull: Optional[bool]
    """Whether to pull punctuation closer."""


class StyleObject(TypedDict):
    """
    The elements of a style string.
    """

    colorize: Optional[bool]
    """Boolean to use standardized colors for symbol groups."""
    padding: Optional[int]
    """Integer value for padding around symbol or sign."""
    background: Optional[str]
    """CSS name or hex color for background."""
    detail: Optional[List[str]]
    """Array for CSS name or hex color for line and optional fill."""
    zoom: Optional[float]
    """Decimal value for zoom level."""
    detailsym: Optional[List[Dict[str, Union[int, List[str]]]]]
    """Custom colors for individual symbols: list of {index: int, detail: List[str]}."""
    classes: Optional[str]
    """List of class names separated with spaces used for SVG."""
    id: Optional[str]
    """ID name used for SVG."""


class ColumnOptions(TypedDict):
    """
    Options for column layout.
    """

    height: Optional[int]
    """The height of the columns."""
    width: Optional[int]
    """The widths of the columns."""
    offset: Optional[int]
    """The lane offset for left and right lanes."""
    pad: Optional[int]
    """Amount of padding before and after signs as well as at top, left, and right of columns."""
    margin: Optional[int]
    """Amount of space at bottom of column that is not available."""
    dynamic: Optional[bool]
    """Enables variable width columns."""
    background: Optional[str]
    """Background color for columns."""
    style: Optional[StyleObject]
    """An object of style options."""
    punctuation: Optional[ColumnPunctuation]
    """An object of punctuation options."""


class SegmentInfo(TypedDict):
    """
    Information about a text segment.
    """

    minX: int
    """The min x value within the segment."""
    minY: int
    """The min y value within the segment."""
    width: int
    """The width of the text segment."""
    height: int
    """The height of the text segment."""
    lane: int
    """Left as -1, Middle as 0, Right as 1."""
    padding: int
    """The padding of the text segment affects colored background."""
    segment: str
    """'sign' or 'symbol'."""
    zoom: float
    """The zoom size of the segment."""


class ColumnSegment(SegmentInfo):
    """A segment in a column layout (extends SegmentInfo with position and text)."""

    x: int
    """The x position in the column."""
    y: int
    """The y position in the column."""
    text: str
    """The text of the sign or symbol with optional style string."""


ColumnData = List[ColumnSegment]

# Symbol and sign types


class SymbolObject(TypedDict):
    """
    The elements of a symbol string.
    """

    symbol: Optional[str]
    """Symbol identifier."""
    coord: Optional[List[int]]
    """X, y coordinate."""
    style: Optional[str]
    """Style string."""


class SignSpatial(TypedDict):
    """A spatial symbol in a sign."""

    symbol: str
    """Symbol identifier."""
    coord: List[int]
    """Coordinate [x, y]."""


class SignObject(TypedDict):
    """
    The elements of a sign string.
    """

    sequence: Optional[List[str]]
    """Array of symbols."""
    box: Optional[str]
    """Signbox marker or lane."""
    max: Optional[List[int]]
    """Preprocessed x, y coordinate."""
    spatials: Optional[List[SignSpatial]]
    """Array of symbols with coordinates."""
    style: Optional[str]
    """Style string."""


__all__ = [
    "ColumnData",
    "ColumnOptions",
    "ColumnPunctuation",
    "ColumnSegment",
    "QueryObject",
    "QueryPrefix",
    "QuerySignboxElement",
    "QuerySignboxOr",
    "QuerySignboxRange",
    "QuerySignboxSymbol",
    "SegmentInfo",
    "SignObject",
    "SignSpatial",
    "StyleObject",
    "SymbolObject",
]



"""
The fsw module contains functions for handling Formal SignWriting in ASCII (FSW) characters.

FSW characters definition: https://datatracker.ietf.org/doc/id/draft-slevinski-formal-signwriting-10.html#name-formal-signwriting-in-ascii
"""

import re
from typing import Any, Dict, List, Optional, Tuple, Union, Set, TypedDict, cast
from dataclasses import dataclass

from .convert import drop_none, fsw_to_coord

from .regex import (
    fsw_pattern_box,
    fsw_pattern_coord,
    fsw_pattern_null_or_symbol,
    fsw_pattern_prefix,
    fsw_pattern_signbox,
    fsw_pattern_spatial,
    fsw_pattern_symbol,
    style_pattern_full,
)

from .style import (
    style_parse,
)

# ----------------------------
# FSW structure constants
# ----------------------------
# Names converted to fsw_structure_*

# kinds: writing, location, punctuation (numbers are the same as JS)
fsw_structure_kind: List[int] = [0x100, 0x37F, 0x387]

# categories: hand, movement, dynamics, head, trunk & limb, location, punctuation.
fsw_structure_category: List[int] = [0x100, 0x205, 0x2F7, 0x2FF, 0x36D, 0x37F, 0x387]

# the 30 symbol groups
fsw_structure_group: List[int] = [
    0x100,
    0x10E,
    0x11E,
    0x144,
    0x14C,
    0x186,
    0x1A4,
    0x1BA,
    0x1CD,
    0x1F5,
    0x205,
    0x216,
    0x22A,
    0x255,
    0x265,
    0x288,
    0x2A6,
    0x2B7,
    0x2D5,
    0x2E3,
    0x2F7,
    0x2FF,
    0x30A,
    0x32A,
    0x33B,
    0x359,
    0x36D,
    0x376,
    0x37F,
    0x387,
]

# ranges: mapping from names to start/end inclusive (same semantics as JS)
fsw_structure_ranges: Dict[str, Tuple[int, int]] = {
    "all": (0x100, 0x38B),
    "writing": (0x100, 0x37E),
    "hand": (0x100, 0x204),
    "movement": (0x205, 0x2F6),
    "dynamic": (0x2F7, 0x2FE),
    "head": (0x2FF, 0x36C),
    "hcenter": (0x2FF, 0x36C),
    "vcenter": (0x2FF, 0x375),
    "trunk": (0x36D, 0x375),
    "limb": (0x376, 0x37E),
    "location": (0x37F, 0x386),
    "punctuation": (0x387, 0x38B),
}


def fsw_is_type(sym_key: str, type_name: str) -> bool:
    """
    Test whether an FSW symbol key is of the given type/range.

    Args:
        sym_key: an FSW symbol key
        type_name: the name of a symbol range

    Returns:
        True if symbol of specified type

    Example:
        >>> fsw_is_type('S10000', 'hand')
        True

    Note:
        The following `type_name` values are supported:

        - **all** - All symbols used in Formal SignWriting.
        - **writing** - Symbols that can be used in the spatial signbox or the temporal prefix.
        - **hand** - Various handshapes.
        - **movement** - Contact symbols, small finger movements, straight arrows, curved arrows, and circles.
        - **dynamic** - Dynamic symbols used to express "feeling" or "tempo" of movement.
        - **head** - Symbols for the head and face.
        - **hcenter** - Used to determine the horizontal center of a sign (same as the head type).
        - **vcenter** - Used to determine the vertical center of a sign (includes head and trunk types).
        - **trunk** - Symbols for torso movement, shoulders, and hips.
        - **limb** - Symbols for limbs and fingers.
        - **location** - Detailed location symbols used only in the temporal prefix.
        - **punctuation** - Symbols used to divide signs into sentences.
    """
    if not (isinstance(sym_key, str) and sym_key.startswith("S") and len(sym_key) >= 4):
        return False
    # the group id is hex in positions 1..3 (slice [1:4])
    try:
        dec = int(sym_key[1:4], 16)
    except ValueError:
        return False
    rng = fsw_structure_ranges.get(type_name)
    if not rng:
        return False
    return rng[0] <= dec <= rng[1]


# ----------------------------
# FSW Colorize
# ----------------------------

fsw_colors: List[str] = [
    "#0000CC",
    "#CC0000",
    "#FF0099",
    "#006600",
    "#000000",
    "#884411",
    "#FF9900",
]


def fsw_colorize(key: str) -> str:
    """
    Function that returns the standardized color for a symbol.

    Args:
        key: an FSW symbol key

    Returns:
        name of standardized color for symbol

    Example:
        >>> fsw_colorize('S10000')
        '#0000CC'
    """
    if not isinstance(key, str):
        return "#000000"
    parsed = fsw_parse_symbol(key)
    color = "#000000"
    if symbol := parsed.get("symbol"):
        dec = int(symbol[1:4], 16)
        index = next(
            (i for i, val in enumerate(fsw_structure_category) if val > dec), -1
        )
        color = fsw_colors[6 if index < 0 else index - 1]
    return color


# ----------------------------
# FSW Parsing
# ----------------------------


def fsw_parse_symbol(fsw_sym: str) -> Dict[str, Any]:
    """
    Parse an FSW symbol with optional coordinate and style string.

    Args:
        fsw_sym: an FSW symbol string

    Returns:
        Dictionary with 'symbol', 'coord', 'style' keys

    Example:
        >>> fsw_parse_symbol('S10000500x500-C')
        {'symbol': 'S10000', 'coord': [500, 500], 'style': '-C'}
    """
    pattern = re.compile(
        rf"^({fsw_pattern_symbol})({fsw_pattern_coord})?({style_pattern_full})?"
    )
    m = pattern.match(fsw_sym)
    if not m:
        return {}
    return drop_none(
        {
            "symbol": m.group(1),
            "coord": fsw_to_coord(m.group(2)) if m.group(2) else None,
            "style": m.group(3),
        }
    )


def fsw_parse_sign(fsw_sign: str) -> Dict[str, Any]:
    """
    Parse an FSW sign with optional style string.

    Args:
        fsw_sign: an FSW sign string

    Returns:
        Dictionary with 'sequence', 'box', 'max', 'spatials', 'style' keys

    Example:
        >>> fsw_parse_sign('AS10011S10019S2e704S2e748M525x535S2e748483x510S10011501x466S2e704510x500S10019476x475-C')
        {'sequence': ['S10011', 'S10019', 'S2e704', 'S2e748'],
         'box': 'M',
         'max': [525, 535],
         'spatials': [{'symbol': 'S2e748', 'coord': [483, 510]},
                      {'symbol': 'S10011', 'coord': [501, 466]},
                      {'symbol': 'S2e704', 'coord': [510, 500]},
                      {'symbol': 'S10019', 'coord': [476, 475]}],
         'style': '-C'}
    """
    pattern = re.compile(
        rf"^({fsw_pattern_prefix})?({fsw_pattern_signbox})({style_pattern_full})?"
    )
    m = pattern.match(fsw_sign)
    if not m:
        return {}
    prefix = m.group(1)
    signbox = m.group(2)
    sequence = (
        re.findall(r".{6}", prefix[1:]) if prefix else None
    )  # skip 'A', every 6 chars
    box = signbox[0]
    max_coord = fsw_to_coord(signbox[1:8])
    spatials_str = signbox[8:] if len(signbox) > 8 else ""
    spatials = (
        [
            {"symbol": s[:6], "coord": fsw_to_coord(s[6:])}
            for s in re.findall(r".{13}", spatials_str)
        ]
        if spatials_str
        else None
    )
    return drop_none(
        {
            "sequence": sequence,
            "box": box,
            "max": max_coord,
            "spatials": spatials,
            "style": m.group(3),
        }
    )


def fsw_parse_text(fsw_text: str) -> List[str]:
    """
    Parse an FSW text into signs and punctuations.

    Args:
        fsw_text: an FSW text string

    Returns:
        List of FSW signs and punctuations

    Example:
        >>> fsw_parse_text('AS14c20S27106M518x529S14c20481x471S27106503x489 AS18701S1870aS2e734S20500M518x533S1870a489x515S18701482x490S20500508x496S2e734500x468 S38800464x496')
        ['AS14c20S27106M518x529S14c20481x471S27106503x489',
         'AS18701S1870aS2e734S20500M518x533S1870a489x515S18701482x490S20500508x496S2e734500x468',
         'S38800464x496']
    """
    if not isinstance(fsw_text, str):
        return []
    pattern = re.compile(
        rf"(?:{fsw_pattern_prefix})?(?:{fsw_pattern_signbox})(?:{style_pattern_full})?|{fsw_pattern_spatial}(?:{style_pattern_full})?"
    )
    return pattern.findall(fsw_text)


# ----------------------------
# FSW Compose
# ----------------------------


def fsw_compose_symbol(fsw_sym_object: Dict[str, Any]) -> Optional[str]:
    """
    Function to compose an fsw symbol with optional coordinate and style string.

    Args:
        fsw_sym_object: an fsw symbol object with keys:
            - 'symbol': str
            - 'coord': Optional[List[int]]
            - 'style': Optional[str]

    Returns:
        an fsw symbol string

    Example:
        >>> fsw_compose_symbol({'symbol': 'S10000', 'coord': [480, 480], 'style': '-C'})
        'S10000480x480-C'
    """
    symbol = fsw_sym_object.get("symbol")
    if not isinstance(symbol, str):
        return None
    symbol_match = re.match(f"^({fsw_pattern_symbol})$", symbol)
    if not symbol_match:
        return None
    symbol = symbol_match.group(1)

    coord = fsw_sym_object.get("coord")
    coord_str = ""
    if coord and isinstance(coord, list) and len(coord) == 2:
        x = str(coord[0])
        y = str(coord[1])
        coord_candidate = f"{x}x{y}"
        coord_match = re.match(f"^({fsw_pattern_coord})$", coord_candidate)
        if coord_match:
            coord_str = coord_match.group(1)

    style_str = fsw_sym_object.get("style")
    style_candidate = ""
    if isinstance(style_str, str):
        style_match = re.match(f"^({style_pattern_full})$", style_str)
        if style_match:
            style_candidate = style_match.group(1)

    return symbol + coord_str + style_candidate


def fsw_compose_sign(fsw_sign_object: Dict[str, Any]) -> Optional[str]:
    """
    Function to compose an fsw sign with style string.

    Args:
        fsw_sign_object: an fsw sign object with keys:
            - 'sequence': Optional[List[str]]
            - 'box': Optional[str]
            - 'max': Optional[List[int]]
            - 'spatials': Optional[List[Dict[str, Tuple[int, int]]]]
            - 'style': Optional[str]

    Returns:
        an fsw sign string

    Example:
        >>> fsw_compose_sign({
        ...     'sequence': ['S10011', 'S10019', 'S2e704', 'S2e748'],
        ...     'box': 'M',
        ...     'max': [525, 535],
        ...     'spatials': [
        ...         {'symbol': 'S2e748', 'coord': [483, 510]},
        ...         {'symbol': 'S10011', 'coord': [501, 466]},
        ...         {'symbol': 'S2e704', 'coord': [510, 500]},
        ...         {'symbol': 'S10019', 'coord': [476, 475]}
        ...     ],
        ...     'style': '-C'
        ... })
        'AS10011S10019S2e704S2e748M525x535S2e748483x510S10011501x466S2e704510x500S10019476x475-C'
    """
    box = fsw_sign_object.get("box", "")
    box_match = re.match(f"^({fsw_pattern_box})$", box)
    if not box_match:
        return None
    box = box_match.group(1)

    max_coord = fsw_sign_object.get("max")
    max_str = ""
    if max_coord and isinstance(max_coord, list) and len(max_coord) == 2:
        x = str(max_coord[0])
        y = str(max_coord[1])
        max_candidate = f"{x}x{y}"
        max_match = re.match(f"^({fsw_pattern_coord})$", max_candidate)
        if max_match:
            max_str = max_match.group(1)
    if not max_str:
        return None

    prefix = ""
    sequence = fsw_sign_object.get("sequence")
    if sequence and isinstance(sequence, list):
        prefix_parts = []
        for key in sequence:
            key_match = re.match(f"^({fsw_pattern_null_or_symbol})$", key)
            if key_match:
                prefix_parts.append(key_match.group(1))
        prefix = "A" + "".join(prefix_parts) if prefix_parts else ""

    signbox = ""
    spatials = fsw_sign_object.get("spatials")
    if spatials and isinstance(spatials, list):
        signbox_parts = []
        for spatial in spatials:
            sym = spatial.get("symbol")
            if not isinstance(sym, str):
                continue
            sym_match = re.match(f"^({fsw_pattern_symbol})$", sym)
            if not sym_match:
                continue
            sym = sym_match.group(1)

            coord = spatial.get("coord")
            coord_str = ""
            if coord and isinstance(coord, list) and len(coord) == 2:
                x = str(coord[0])
                y = str(coord[1])
                coord_candidate = f"{x}x{y}"
                coord_match = re.match(f"^({fsw_pattern_coord})$", coord_candidate)
                if coord_match:
                    coord_str = coord_match.group(1)
            if not coord_str:
                continue

            signbox_parts.append(sym + coord_str)
        signbox = "".join(signbox_parts)

    style_str = fsw_sign_object.get("style")
    style_candidate = ""
    if isinstance(style_str, str):
        style_match = re.match(f"^({style_pattern_full})$", style_str)
        if style_match:
            style_candidate = style_match.group(1)

    return prefix + box + max_str + signbox + style_candidate


# ----------------------------
# FSW Info
# ----------------------------


def fsw_info(fsw: str) -> Dict[str, Any]:
    """
    Function to gather sizing information about an fsw sign or symbol.

    Args:
        fsw: an fsw sign or symbol

    Returns:
        information about the fsw string

    Example:
        >>> fsw_info('AS14c20S27106L518x529S14c20481x471S27106503x489-P10Z2')
        {
            'minX': 481,
            'minY': 471,
            'width': 37,
            'height': 58,
            'lane': -1,
            'padding': 10,
            'segment': 'sign',
            'zoom': 2
        }
    """
    lanes = {"B": 0, "L": -1, "M": 0, "R": 1}
    parsed_sign = fsw_parse_sign(fsw)
    width: Optional[int] = None
    height: Optional[int] = None
    segment: str = "none"
    x1: int = 490
    y1: int = 490
    lane: int = 0
    parsed: Dict[str, Any] = {}  # Unified parsed object for style

    if spatials := parsed_sign.get("spatials"):
        x_coords = [spatial["coord"][0] for spatial in spatials]
        y_coords = [spatial["coord"][1] for spatial in spatials]
        x1 = min(x_coords)
        y1 = min(y_coords)
        x2, y2 = parsed_sign["max"]
        width = x2 - x1
        height = y2 - y1
        segment = "sign"
        lane = lanes[parsed_sign["box"]]
        parsed = parsed_sign
    else:
        parsed_symbol = fsw_parse_symbol(fsw)
        lane = lanes["M"]
        if coord := parsed_symbol.get("coord"):
            x1, y1 = coord
            width = (500 - x1) * 2
            height = (500 - y1) * 2
            segment = "symbol"
        parsed = parsed_symbol

    style = style_parse(parsed.get("style", ""))
    zoom = style.get("zoom", 1)
    padding = style.get("padding", 0)

    return {
        "minX": x1,
        "minY": y1,
        "width": width or 20,
        "height": height or 20,
        "segment": segment,
        "lane": lane,
        "padding": padding,
        "zoom": zoom,
    }


# ----------------------------
# FSW Columns
# ----------------------------

fsw_column_defaults: Dict[str, Any] = {
    "height": 500,
    "width": 150,
    "offset": 50,
    "pad": 20,
    "margin": 5,
    "dynamic": False,
    "background": None,
    "punctuation": {
        "spacing": True,
        "pad": 30,
        "pull": True,
    },
    "style": {
        "detail": ["black", "white"],
        "zoom": 1,
    },
}


def fsw_column_defaults_merge(
    options: Optional[Dict[str, Any]] = None
) -> Dict[str, Any]:
    """
    Function to merge an object of column options with default values.

    Args:
        options: object of column options

    Returns:
        object of column options merged with column defaults

    Example:
        >>> fsw_column_defaults_merge({'height': 500, 'width': 150})
        {'height': 500, 'width': 150, 'offset': 50, ...}
    """
    if not isinstance(options, dict):
        options = {}
    merged = {**fsw_column_defaults, **options}
    merged["punctuation"] = {
        **fsw_column_defaults["punctuation"],
        **(options.get("punctuation") or {}),
    }
    merged["style"] = {**fsw_column_defaults["style"], **(options.get("style") or {})}
    return drop_none(merged)


def fsw_columns(
    fsw_text: str, options: Optional[Dict[str, Any]] = None
) -> Dict[str, Any]:
    """
    Function to transform an FSW text to an array of columns.

    Args:
        fsw_text: FSW text of signs and punctuation
        options: object of column options

    Returns:
        object of column options, widths array, and column data

    Example:
        >>> fsw_columns('AS14c20S27106M518x529S14c20481x471S27106503x489 AS18701S1870aS2e734S20500M518x533S1870a489x515S18701482x490S20500508x496S2e734500x468 S38800464x496', {'height': 500, 'width': 150})
        {'options': {...}, 'widths': [150], 'columns': [[{'x': 56, 'y': 20, ...}, ...]]}
    """
    if not isinstance(fsw_text, str):
        return {}
    values = fsw_column_defaults_merge(options)
    input_text = fsw_parse_text(fsw_text)
    if not input_text:
        return {}

    cursor = 0
    cols: List[List[Dict[str, Any]]] = []
    col: List[Dict[str, Any]] = []
    plus = 0
    center = values["width"] // 2
    max_height = values["height"] - values["margin"]
    pullable = True
    finalize = False

    for val in input_text:
        informed = fsw_info(val)
        cursor += plus
        if values["punctuation"]["spacing"]:
            cursor += values["pad"] if informed["segment"] == "sign" else 0
        else:
            cursor += values["pad"]

        finalize = (cursor + informed["height"]) > max_height

        if (
            informed["segment"] == "symbol"
            and values["punctuation"]["pull"]
            and pullable
        ):
            finalize = False
            pullable = False

        if not col:
            finalize = False

        if finalize:
            cursor = values["pad"]
            cols.append(col)
            col = []
            pullable = True

        item = {**informed, "text": val}
        item["x"] = (
            center
            + (values["offset"] * informed["lane"])
            - ((500 - informed["minX"]) * informed["zoom"] * values["style"]["zoom"])
        )
        item["y"] = cursor
        col.append(item)
        cursor += informed["height"] * informed["zoom"] * values["style"]["zoom"]

        if values["punctuation"]["spacing"]:
            plus = (
                values["pad"]
                if informed["segment"] == "sign"
                else values["punctuation"]["pad"]
            )
        else:
            plus = values["pad"]

    if col:
        cols.append(col)

    # Over-height adjustment for pulled punctuation
    if values["punctuation"]["pull"]:
        for c in cols:
            last = c[-1]
            diff = (last["y"] + last["height"]) - (values["height"] - values["margin"])
            if diff > 0:
                adj = (diff // len(c)) + 1
                for i, item in enumerate(c):
                    item["y"] -= adj * i + adj

    # Contract, expand, adjust widths
    widths: List[int] = []
    for c in cols:
        mins = [center - values["offset"] - values["pad"]]
        maxs = [center + values["offset"] + values["pad"]]
        for item in c:
            mins.append(item["x"] - values["pad"])
            maxs.append(item["x"] + item["width"] + values["pad"])
        min_val = min(mins)
        max_val = max(maxs)

        width = values["width"]
        adj = 0
        if not values["dynamic"]:
            adj = center - ((min_val + max_val) // 2)
        else:
            width = max_val - min_val
            adj = -min_val

        for item in c:
            item["x"] += adj
        widths.append(width)

    return {"options": values, "widths": widths, "columns": cols}


# ----------------------------
# FSW Tokenizer
# ----------------------------


class SpecialToken(TypedDict):
    index: int
    name: str
    value: str


class TokenizerMappings(TypedDict):
    i2s: Dict[int, str]
    s2i: Dict[str, int]


DEFAULT_SPECIAL_TOKENS: List[SpecialToken] = [
    {"index": 0, "name": "UNK", "value": "[UNK]"},
    {"index": 1, "name": "PAD", "value": "[PAD]"},
    {"index": 2, "name": "CLS", "value": "[CLS]"},
    {"index": 3, "name": "SEP", "value": "[SEP]"},
]


def _generate_tokens() -> List[str]:
    sequence = ["A"]
    signbox = ["B", "L", "M", "R"]
    null_token = ["S000"]
    base_symbols = [f"S{i:03x}" for i in range(0x100, 0x38C)]
    rows = [f"r{i:x}" for i in range(16)]
    cols = [f"c{i:x}" for i in range(6)]
    positions = [f"p{i}" for i in range(250, 751)]
    return sequence + signbox + null_token + base_symbols + rows + cols + positions


@dataclass
class SpecialTokenMappings:
    by_index: Dict[int, SpecialToken]
    by_name: Dict[str, SpecialToken]
    by_value: Dict[str, SpecialToken]

    def get_by_index(self, index: int) -> SpecialToken:
        unk = next(t for t in self.by_index.values() if t["name"] == "UNK")
        return self.by_index.get(index, unk)

    def get_by_name(self, name: str) -> SpecialToken:
        return self.by_name.get(name, self.by_name["UNK"])

    def get_by_value(self, value: str) -> SpecialToken:
        return self.by_value.get(value, self.by_name["UNK"])

    def get_all_values(self) -> List[str]:
        return [t["value"] for t in self.by_index.values()]

    def get_all_indices(self) -> List[int]:
        return [t["index"] for t in self.by_index.values()]


def _create_special_token_mappings(
    special_tokens: List[SpecialToken],
) -> SpecialTokenMappings:
    by_index: Dict[int, SpecialToken] = {}
    by_name: Dict[str, SpecialToken] = {}
    by_value: Dict[str, SpecialToken] = {}
    indices: Set[int] = set()

    for token in special_tokens:
        if token["index"] in indices:
            raise ValueError(f"Duplicate token index: {token['index']}")
        indices.add(token["index"])
        by_index[token["index"]] = token
        by_name[token["name"]] = token
        by_value[token["value"]] = token

    return SpecialTokenMappings(by_index, by_name, by_value)


def _create_token_mappings(
    tokens: List[str],
    special_token_mappings: SpecialTokenMappings,
    starting_index: int,
) -> TokenizerMappings:
    i2s: Dict[int, str] = {}
    s2i: Dict[str, int] = {}

    # Add special tokens first
    for special_token in special_token_mappings.by_index.values():
        i2s[special_token["index"]] = special_token["value"]
        s2i[special_token["value"]] = special_token["index"]

    # Add regular tokens
    for i, reg_token in enumerate(tokens):
        index = starting_index + i
        i2s[index] = reg_token
        s2i[reg_token] = index

    return {"i2s": i2s, "s2i": s2i}


def fsw_tokenize(
    fsw: str,
    sequence: bool = True,
    signbox: bool = True,
    sep: Optional[str] = "[SEP]",
) -> List[str]:
    """
    Tokenizes an FSW string into an array of tokens

    Args:
        fsw: FSW string to tokenize
        sequence: Whether to include sequence tokens
        signbox: Whether to include signbox tokens
        sep: Separator token

    Returns:
        Array of tokens

    Example:
        >>> fsw_tokenize("AS10e00M507x515S10e00492x485", sequence=False, sep=None)
        ['M', 'p507', 'p515', 'S10e', 'c0', 'r0', 'p492', 'p485']
    """

    def tokenize_symbol(symbol: str) -> List[str]:
        return [symbol[:4], f"c{symbol[4]}", f"r{symbol[5]}"]

    def tokenize_coord(coord: List[int]) -> List[str]:
        return [f"p{c}" for c in coord]

    segments = fsw_parse_text(fsw)
    result: List[str] = []

    for fsw_segment in segments:
        tokens: List[str] = []

        if re.search(r"[BLMR]", fsw_segment):
            sign = fsw_parse_sign(fsw_segment)

            if sign.get("sequence") and sequence:
                tokens.extend(["A"])
                for seq_item in sign["sequence"]:
                    tokens.extend(tokenize_symbol(seq_item))

            if signbox:
                tokens.append(sign["box"])
                tokens.extend(tokenize_coord(sign["max"]))
                for spatial in sign.get("spatials", []):
                    tokens.extend(tokenize_symbol(spatial["symbol"]))
                    tokens.extend(tokenize_coord(spatial["coord"]))

        else:
            parsed = fsw_parse_symbol(fsw_segment)

            if not signbox and not sequence:
                continue

            if not signbox and sequence:
                tokens = ["A"] + tokenize_symbol(parsed["symbol"])
            else:
                coord = parsed.get("coord")
                if coord:
                    rev_coord = [1000 - c for c in coord]
                    tokens = (
                        ["M"]
                        + tokenize_coord(rev_coord)
                        + tokenize_symbol(parsed["symbol"])
                        + tokenize_coord(coord)
                    )

        if tokens and sep:
            tokens.append(sep)
        result.extend(tokens)

    return result


def fsw_detokenize(
    tokens: List[str],
    special_tokens: List[SpecialToken] = DEFAULT_SPECIAL_TOKENS,
) -> str:
    """
    Converts an array of tokens back into an FSW string

    Args:
        tokens: Array of tokens to convert
        special_tokens: Array of special token objects to filter out

    Returns:
        FSW string

    Example:
        >>> fsw_detokenize(['M', 'p507', 'p515', 'S10e', 'c0', 'r0', 'p492', 'p485'])
        "M507x515S10e00492x485"
    """
    special_values = {t["value"] for t in special_tokens}
    filtered = [t for t in tokens if t not in special_values]
    joined = " ".join(filtered)
    joined = re.sub(r"p(\d{3})\s+p(\d{3})", r"\1x\2", joined)
    joined = re.sub(r" c(\d)\d? r(.)", r"\1\2", joined)
    joined = re.sub(r" c(\d)\d?", r"\g<1>0", joined)
    joined = re.sub(r" r(.)", r"0\1", joined)
    joined = joined.replace(" ", "")
    joined = re.sub(r"(\d)([BLMR])", r"\1 \2", joined)
    joined = re.sub(r"(\d)(AS)", r"\1 \2", joined)
    joined = re.sub(
        r"(A(?:S00000|S[123][0-9a-f]{2}[0-5][0-9a-f])+)( )([BLMR])", r"\1\3", joined
    )
    return joined


def fsw_chunk_tokens(
    tokens: List[str],
    chunk_size: int,
    cls: str = "[CLS]",
    sep: str = "[SEP]",
    pad: str = "[PAD]",
) -> List[List[str]]:
    """
    Splits tokens into chunks of specified size while preserving sign boundaries

    Args:
        tokens: Array of tokens to chunk
        chunk_size: Maximum size of each chunk
        cls: CLS token
        sep: SEP token
        pad: PAD token

    Returns:
        Array of token chunks
    """
    if chunk_size < 60:
        raise ValueError(
            "Chunk size must be at least 60 tokens to accommodate a typical sign"
        )

    chunks: List[List[str]] = []
    token_index = 0

    while token_index < len(tokens):
        current_chunk = [cls]

        while token_index < len(tokens):
            look_ahead = token_index
            while look_ahead < len(tokens) and tokens[look_ahead] != sep:
                look_ahead += 1

            sign_size = look_ahead - token_index + 1

            if len(current_chunk) + sign_size > chunk_size - 1:
                break

            current_chunk.extend(tokens[token_index : look_ahead + 1])
            token_index = look_ahead + 1

        while len(current_chunk) < chunk_size:
            current_chunk.append(pad)

        chunks.append(current_chunk)

    return chunks


class FSWTokenizer:
    """
    Creates a tokenizer object with encoding and decoding capabilities.

    Args:
        special_tokens: Special tokens list
        starting_index: Starting index for regular tokens

    Example:
        >>> t = FSWTokenizer()
        >>> t.encode('M507x515S10e00492x485')
        [7, 941, 949, 24, 678, 662, 926, 919, 3]
    """

    def __init__(
        self,
        special_tokens: List[SpecialToken] = DEFAULT_SPECIAL_TOKENS,
        starting_index: Optional[int] = None,
    ):
        self.special_tokens = _create_special_token_mappings(special_tokens)
        max_special = (
            max(self.special_tokens.get_all_indices())
            if self.special_tokens.get_all_indices()
            else -1
        )
        calculated_starting_index = starting_index or (max_special + 1)
        tokens = _generate_tokens()
        mappings = _create_token_mappings(
            tokens, self.special_tokens, calculated_starting_index
        )
        self.i2s: Dict[int, str] = mappings["i2s"]
        self.s2i: Dict[str, int] = mappings["s2i"]
        self.length = len(self.i2s)

    def vocab(self) -> List[str]:
        return list(self.i2s.values())

    def encode_tokens(self, tokens: List[str]) -> List[int]:
        return [
            self.s2i.get(t, self.special_tokens.get_by_value(t)["index"])
            for t in tokens
        ]

    def decode_tokens(self, indices: List[int]) -> List[str]:
        unk_value = self.special_tokens.get_by_name("UNK")["value"]
        return [self.i2s.get(i, unk_value) for i in indices]

    def encode(
        self, text: str, sequence: bool = True, signbox: bool = True
    ) -> List[int]:
        sep_value = self.special_tokens.get_by_name("SEP")["value"]
        tokens = fsw_tokenize(text, sequence=sequence, signbox=signbox, sep=sep_value)
        return self.encode_tokens(tokens)

    def decode(self, tokens: Union[List[int], List[List[int]]]) -> str:
        if not tokens:
            return ""

        if isinstance(tokens[0], list):
            decoded_chunks = [
                self.decode_tokens(cast(List[int], chunk)) for chunk in tokens
            ]
            flat_tokens = [t for chunk in decoded_chunks for t in chunk]
            return fsw_detokenize(
                flat_tokens, list(self.special_tokens.by_index.values())
            )

        decoded_tokens = self.decode_tokens(cast(List[int], tokens))
        return fsw_detokenize(
            decoded_tokens, list(self.special_tokens.by_index.values())
        )

    def chunk(self, tokens: List[str], chunk_size: int) -> List[List[str]]:
        return fsw_chunk_tokens(
            tokens,
            chunk_size,
            cls=self.special_tokens.get_by_name("CLS")["value"],
            sep=self.special_tokens.get_by_name("SEP")["value"],
            pad=self.special_tokens.get_by_name("PAD")["value"],
        )


__all__ = [
    "fsw_structure_kind",
    "fsw_structure_category",
    "fsw_structure_group",
    "fsw_structure_ranges",
    "fsw_is_type",
    "fsw_colors",
    "fsw_colorize",
    "fsw_parse_symbol",
    "fsw_parse_sign",
    "fsw_parse_text",
    "fsw_compose_symbol",
    "fsw_compose_sign",
    "fsw_info",
    "fsw_column_defaults",
    "fsw_column_defaults_merge",
    "fsw_columns",
    "fsw_tokenize",
    "fsw_detokenize",
    "fsw_chunk_tokens",
    "FSWTokenizer",
]



"""
The style module contains regular expressions and functions for parsing and composing style strings.

Style string definition: https://datatracker.ietf.org/doc/id/draft-slevinski-formal-signwriting-10.html#name-styling-string
"""

import re
from typing import Any, Dict, List, Optional, Union

from .regex import (
    style_pattern_colorbase,
    style_pattern_detailsym,
    style_pattern_full_groups,
    style_pattern_full,
)

# ----------------------------
# Style Parsing
# ----------------------------


def _prefix_color(color: str) -> str:
    if re.match(rf"^{style_pattern_colorbase}$", color):
        return (
            "#" + color
            if len(color) in (3, 6)
            and all(c in "0123456789abcdefABCDEF" for c in color)
            else color
        )
    return color


def style_parse(style_string: str) -> Dict[str, Any]:
    """
    Parse a style string to a structured dictionary.

    Args:
        style_string: a style string

    Returns:
        Dictionary with style elements

    Example:
        >>> style_parse('-CP10G_blue_D_red,Cyan_')
        {'colorize': True, 'padding': 10, 'background': 'blue', 'detail': ['red', 'Cyan']}
    """
    pattern = re.compile(rf"^{style_pattern_full_groups}")
    m = pattern.match(style_string) if isinstance(style_string, str) else None
    if not m:
        return {}
    colorize = True if m.group(1) else None
    padding = int(m.group(2)[1:]) if m.group(2) else None
    background = None if not m.group(3) else _prefix_color(m.group(3)[2:-1])
    detail = (
        None
        if not (s := m.group(4))
        else [_prefix_color(p.strip()) for p in s[2:-1].split(",")]
    )
    zoom_str = m.group(5)
    zoom: Optional[Union[float, str]] = None
    if zoom_str:
        zoom = "x" if zoom_str == "Zx" else float(zoom_str[1:])
    detailsym_str = m.group(6)
    detailsym: Optional[List[Dict[str, Any]]] = None
    if detailsym_str:
        detailsym_matches = re.findall(style_pattern_detailsym, detailsym_str)
        detailsym = []
        for ds in detailsym_matches:
            index = int(ds[1:3])
            inner = ds[4:-1]  # strip "D01_" and trailing "_"
            det = [_prefix_color(p.strip()) for p in inner.split(",") if p]
            detailsym.append({"index": index, "detail": det})
    classes = m.group(7) if m.group(7) else None
    id_ = m.group(8) if m.group(8) else None
    res = {
        "colorize": colorize,
        "padding": padding,
        "background": background,
        "detail": detail,
        "zoom": zoom,
        "detailsym": detailsym,
        "classes": classes,
        "id": id_,
    }
    return {k: v for k, v in res.items() if v is not None}


# ----------------------------
# Style Compose
# ----------------------------


def style_compose(style_dict: Dict[str, Any]) -> Optional[str]:
    """
    Function to compose style string from dictionary.

    Args:
        style_dict: a dictionary of style elements with keys:
            - 'colorize': Optional[bool]
            - 'padding': Optional[int]
            - 'background': Optional[str]
            - 'detail': Optional[List[str]]
            - 'zoom': Optional[Union[float, str]]
            - 'detailsym': Optional[List[Dict]] (each with 'index': int, 'detail': List[str])
            - 'classes': Optional[str]
            - 'id': Optional[str]

    Returns:
        style string

    Example:
        >>> style_compose({
        ...     'colorize': True,
        ...     'padding': 10,
        ...     'background': 'blue',
        ...     'detail': ['red', 'Cyan'],
        ...     'zoom': 1.1,
        ...     'detailsym': [
        ...         {'index': 1, 'detail': ['#ff0', 'green']}
        ...     ]
        ... })
        '-CP10GblueDred,CyanZ1.1-D01#ff0,green_'
    """
    style1 = ""
    if style_dict.get("colorize"):
        style1 += "C"
    padding = style_dict.get("padding")
    if padding is not None:
        style1 += f"P{int(padding):02d}"
    background = style_dict.get("background")
    if background:
        style1 += f"G_{background.lstrip('#')}_"
    detail = style_dict.get("detail")
    if detail and isinstance(detail, list):
        style1 += f"D_{','.join(detail).lstrip('#')}_"
    zoom = style_dict.get("zoom")
    if zoom is not None:
        style1 += f"Z{zoom}"

    style2 = ""
    detailsym = style_dict.get("detailsym")
    if detailsym and isinstance(detailsym, list):
        detailsym_parts = []
        for ds in detailsym:
            index = ds.get("index")
            ds_detail = ds.get("detail")
            if index is not None and ds_detail and isinstance(ds_detail, list):
                detailsym_parts.append(
                    f"D{int(index):02d}_{','.join(ds_detail).lstrip('#')}_"
                )
        style2 += "".join(detailsym_parts)

    style3 = ""
    classes = style_dict.get("classes")
    if isinstance(classes, str):
        classes_match = re.match(f"^({style_dict.get('classes')})$", classes)
        if classes_match:
            style3 += classes_match.group(1)
    id_ = style_dict.get("id")
    if classes or id_:
        style3 += "!"
    if id_ and isinstance(id_, str):
        id_match = re.match(f"^({style_dict.get('id')})$", id_)
        if id_match:
            style3 += id_match.group(1) + "!"

    result = (
        "-"
        + style1
        + ("-" + style2 if style2 or style3 else "")
        + ("-" + style3 if style3 else "")
    )
    match = re.match(f"^({style_pattern_full})$", result)
    return match.group(1) if match else None


# ----------------------------
# Style RGB
# ----------------------------


def _rgb_to_arr(rgb: str) -> List[float]:
    if not isinstance(rgb, str):
        return [0.0, 0.0, 0.0]
    m = re.match(r"rgba?\((.+?)\)", rgb.lower())
    if not m:
        return [0.0, 0.0, 0.0]
    values = m.group(1).split(",")
    return [float(v.strip()) for v in values]


def _arr_to_hex(arr: List[float]) -> str:
    return "".join(f"{int(v):02x}" for v in arr[:3])


def style_rgb_to_hex(rgb: str, tolerance: float = 0.0) -> str:
    """
    Function to convert rgb color to hex or "transparent" if below tolerance.

    Args:
        rgb: an rgb color
        tolerance: max alpha for full transparency (default: 0)

    Returns:
        a hex color or "transparent"

    Examples:
        >>> style_rgb_to_hex("rgb(255,255,255)")
        'ffffff'
        >>> style_rgb_to_hex("rgba(255,255,255,0.5)",0.5)
        'transparent'
    """
    arr = _rgb_to_arr(rgb)
    if len(arr) == 4 and arr[3] <= tolerance:
        return "transparent"
    else:
        return _arr_to_hex(arr)


def style_rgba_to_hex(color: str, background: str) -> str:
    """
    Function to merge color with background based on alpha transparency

    Args:
        color: an rgba color
        background: an rgba background color

    Returns:
        a hex color or "transparent"

    Example:
        >>> style_rgba_to_hex("rgba(255,255,255,0.5)","rgb(0,0,0)")
        '7f7f7f'
    """
    b_arr = _rgb_to_arr(background)
    c_arr = _rgb_to_arr(color)
    alpha = c_arr[3] if len(c_arr) == 4 else 1.0
    if alpha == 0:
        return "transparent"
    else:
        merged = [(1 - alpha) * b_arr[i] + alpha * c_arr[i] for i in range(3)]
        return _arr_to_hex(merged)


# ----------------------------
# Style Merge
# ----------------------------


def style_merge(
    style1: Optional[Dict[str, Any]], style2: Optional[Dict[str, Any]]
) -> Dict[str, Any]:
    """
    Function to merge style dictionaries.

    Args:
        style1: a style dictionary
        style2: a style dictionary

    Returns:
        a style dictionary

    Example:
        >>> style_merge({'colorize': True}, {'zoom': 2})
        {'colorize': True, 'zoom': 2}
    """
    if not isinstance(style1, dict):
        style1 = {}
    if not isinstance(style2, dict):
        style2 = {}
    zoom1 = style1.get("zoom", 1)
    zoom2 = style2.get("zoom", 1)
    merged = {**style1, **style2}
    merged["zoom"] = zoom1 * zoom2
    return merged


__all__ = [
    "style_parse",
    "style_compose",
    "style_rgb_to_hex",
    "style_rgba_to_hex",
    "style_merge",
]



"""
The swu module contains functions for handling SignWriting in Unicode (SWU) characters.

SWU characters definition: https://datatracker.ietf.org/doc/id/draft-slevinski-formal-signwriting-10.html#name-signwriting-in-unicode-swu

"""

import re
from typing import Any, Dict, List, Optional, Tuple

from .convert import drop_none, swu_to_coord, coord_to_swu, swu_to_code

from .regex import (
    style_pattern_full,
    swu_pattern_box,
    swu_pattern_null_or_symbol,
    swu_pattern_coord,
    swu_pattern_prefix,
    swu_pattern_signbox,
    swu_pattern_spatial,
    swu_pattern_symbol,
)

from .style import (
    style_parse,
)

# ----------------------------
# SWU structure constants (plane 4 code points)
# ----------------------------
swu_structure_kind: List[int] = [0x40001, 0x4EFA1, 0x4F2A1]

swu_structure_category: List[int] = [
    0x40001,
    0x461E1,
    0x4BCA1,
    0x4BFA1,
    0x4E8E1,
    0x4EFA1,
    0x4F2A1,
]

swu_structure_group: List[int] = [
    0x40001,
    0x40541,
    0x40B41,
    0x41981,
    0x41C81,
    0x43241,
    0x43D81,
    0x445C1,
    0x44CE1,
    0x45BE1,
    0x461E1,
    0x46841,
    0x46FC1,
    0x47FE1,
    0x485E1,
    0x49301,
    0x49E41,
    0x4A4A1,
    0x4AFE1,
    0x4B521,
    0x4BCA1,
    0x4BFA1,
    0x4C3C1,
    0x4CFC1,
    0x4D621,
    0x4E161,
    0x4E8E1,
    0x4EC41,
    0x4EFA1,
    0x4F2A1,
]

swu_structure_ranges: Dict[str, Tuple[int, int]] = {
    "all": (0x40001, 0x4F480),
    "writing": (0x40001, 0x4EFA0),
    "hand": (0x40001, 0x461E0),
    "movement": (0x461E1, 0x4BCA0),
    "dynamic": (0x4BCA1, 0x4BFA0),
    "head": (0x4BFA1, 0x4E8E0),
    "hcenter": (0x4BFA1, 0x4E8E0),
    "vcenter": (0x4BFA1, 0x4EC40),
    "trunk": (0x4E8E1, 0x4EC40),
    "limb": (0x4EC41, 0x4EFA0),
    "location": (0x4EFA1, 0x4F2A0),
    "punctuation": (0x4F2A1, 0x4F480),
}


def swu_is_type(swu_char: str, type_name: str) -> bool:
    """
    Test whether an SWU character is of the given type/range.

    Args:
        swu_char: an SWU symbol
        type_name: the name of a symbol range

    Returns:
        True if symbol of specified type

    Example:
        >>> swu_is_type('񀀁', 'hand')
        True

    Note:
        The following `type_name` values are supported:

        - **all** - All symbols used in Formal SignWriting.
        - **writing** - Symbols that can be used in the spatial signbox or the temporal prefix.
        - **hand** - Various handshapes.
        - **movement** - Contact symbols, small finger movements, straight arrows, curved arrows, and circles.
        - **dynamic** - Dynamic symbols used to express "feeling" or "tempo" of movement.
        - **head** - Symbols for the head and face.
        - **hcenter** - Used to determine the horizontal center of a sign (same as the head type).
        - **vcenter** - Used to determine the vertical center of a sign (includes head and trunk types).
        - **trunk** - Symbols for torso movement, shoulders, and hips.
        - **limb** - Symbols for limbs and fingers.
        - **location** - Detailed location symbols used only in the temporal prefix.
        - **punctuation** - Symbols used to divide signs into sentences.
    """
    if not (isinstance(swu_char, str) and len(swu_char) >= 1):
        return False
    code = ord(swu_char[0])
    rng = swu_structure_ranges.get(type_name)
    if not rng:
        return False
    return rng[0] <= code <= rng[1]


# ----------------------------
# SWU Colorize
# ----------------------------

swu_colors: List[str] = [
    "#0000CC",
    "#CC0000",
    "#FF0099",
    "#006600",
    "#000000",
    "#884411",
    "#FF9900",
]


def swu_colorize(sym: str) -> str:
    """
    Function that returns the standardized color for a symbol.

    Args:
        sym: an SWU symbol

    Returns:
        name of standardized color for symbol

    Example:
        >>> swu_colorize('񀀁')
        '#0000CC'
    """
    if not isinstance(sym, str):
        return "#000000"
    parsed = swu_parse_symbol(sym)
    color = "#000000"
    if symbol := parsed.get("symbol"):
        code = swu_to_code(symbol)
        index = next(
            (i for i, val in enumerate(swu_structure_category) if val > code), -1
        )
        color = swu_colors[6 if index < 0 else index - 1]
    return color


# ----------------------------
# SWU Parsing
# ----------------------------


def swu_parse_symbol(swu_sym: str) -> Dict[str, Any]:
    """
    Parse an SWU symbol with optional coordinate and style string.

    Args:
        swu_sym: an SWU symbol string

    Returns:
        Dictionary with 'symbol', 'coord', 'style' keys


    Example:
        >>> swu_parse_symbol('񀀁𝤆𝤆-C')
        {'symbol': '񀀁', 'coord': [500, 500], 'style': '-C'}
    """
    pattern = re.compile(
        rf"^({swu_pattern_symbol})({swu_pattern_coord})?({style_pattern_full})?"
    )
    m = pattern.match(swu_sym)
    if not m:
        return {}
    return drop_none(
        {
            "symbol": m.group(1),
            "coord": swu_to_coord(m.group(2)) if m.group(2) else None,
            "style": m.group(3) if m.group(3) else None,
        }
    )


def swu_parse_sign(swu_sign: str) -> Dict[str, Any]:
    """
    Parse an SWU sign with optional style string.

    Args:
        swu_sign: an SWU sign string

    Returns:
        Dictionary with 'sequence', 'box', 'max', 'spatials', 'style' keys

    Example:
        >>> swu_parse_sign('𝠀񀀒񀀚񋚥񋛩𝠃𝤟𝤩񋛩𝣵𝤐񀀒𝤇𝣤񋚥𝤐𝤆񀀚𝣮𝣭-C')
        {'sequence': ['񀀒', '񀀚', '񋚥', '񋛩'],
         'box': '𝠃',
         'max': [525, 535],
         'spatials': [{'symbol': '񋛩', 'coord': [483, 510]},
                      {'symbol': '񀀒', 'coord': [501, 466]},
                      {'symbol': '񋚥', 'coord': [510, 500]},
                      {'symbol': '񀀚', 'coord': [476, 475]}],
         'style': '-C'}
    """
    pattern = re.compile(
        rf"^({swu_pattern_prefix})?({swu_pattern_signbox})({style_pattern_full})?"
    )
    m = pattern.match(swu_sign)
    if not m:
        return {}
    prefix = m.group(1)
    signbox = m.group(2)
    sequence = list(prefix[1:]) if prefix else None  # skip 𝠀, each char is a symbol
    box = signbox[0]
    max_coord = swu_to_coord(signbox[1:3])
    spatials_str = signbox[3:] if len(signbox) > 3 else ""
    spatials_matches = re.findall(
        rf"({swu_pattern_symbol})({swu_pattern_coord})", spatials_str
    )
    spatials = (
        [
            {"symbol": sym, "coord": swu_to_coord(coord_str)}
            for sym, coord_str in spatials_matches
        ]
        if spatials_matches
        else None
    )
    return drop_none(
        {
            "sequence": sequence,
            "box": box,
            "max": max_coord,
            "spatials": spatials,
            "style": m.group(3),
        }
    )


def swu_parse_text(swu_text: str) -> List[str]:
    """
    Parse an SWU text into signs and punctuations.

    Args:
        swu_text: an SWU text string

    Returns:
        List of SWU signs and punctuations

    Example:
        >>> swu_parse_text('𝠀񁲡񈩧𝠃𝤘𝤣񁲡𝣳𝣩񈩧𝤉𝣻 𝠀񃊢񃊫񋛕񆇡𝠃𝤘𝤧񃊫𝣻𝤕񃊢𝣴𝣼񆇡𝤎𝤂񋛕𝤆𝣦 񏌁𝣢𝤂')
        ['𝠀񁲡񈩧𝠃𝤘𝤣񁲡𝣳𝣩񈩧𝤉𝣻',
         '𝠀񃊢񃊫񋛕񆇡𝠃𝤘𝤧񃊫𝣻𝤕񃊢𝣴𝣼񆇡𝤎𝤂񋛕𝤆𝣦',
         '񏌁𝣢𝤂']
    """
    if not isinstance(swu_text, str):
        return []
    pattern = re.compile(
        rf"(?:{swu_pattern_prefix})?(?:{swu_pattern_signbox})(?:{style_pattern_full})?|{swu_pattern_spatial}(?:{style_pattern_full})?"
    )
    return pattern.findall(swu_text)


# ----------------------------
# SWU Compose
# ----------------------------


def swu_compose_symbol(swu_sym_object: Dict[str, Any]) -> Optional[str]:
    """
    Function to compose an swu symbol with optional coordinate and style string.

    Args:
        swu_sym_object: an swu symbol object with keys:
            - 'symbol': str
            - 'coord': Optional[List[int]]
            - 'style': Optional[str]

    Returns:
        an swu symbol string

    Example:
        >>> swu_compose_symbol({'symbol': '񀀁', 'coord': [500, 500], 'style': '-C'})
        '񀀁𝤆𝤆-C'
    """
    if not isinstance(swu_sym_object, dict):
        return None
    symbol = swu_sym_object.get("symbol")
    if not isinstance(symbol, str):
        return None
    symbol_match = re.match(f"^({swu_pattern_symbol})$", symbol)
    if not symbol_match:
        return None
    symbol = symbol_match.group(1)

    coord = swu_sym_object.get("coord")
    coord_str = ""
    if coord and isinstance(coord, list) and len(coord) == 2:
        coord_str = coord_to_swu(coord)

    style_str = swu_sym_object.get("style")
    style_candidate = ""
    if isinstance(style_str, str):
        style_match = re.match(f"^({style_pattern_full})$", style_str)
        if style_match:
            style_candidate = style_match.group(1)

    return symbol + coord_str + style_candidate


def swu_compose_sign(swu_sign_object: Dict[str, Any]) -> Optional[str]:
    """
    Function to compose an swu sign with style string.

    Args:
        swu_sign_object: an swu sign object with keys:
            - 'sequence': Optional[List[str]]
            - 'box': Optional[str]
            - 'max': Optional[List[int]]
            - 'spatials': Optional[List[Dict[str, Tuple[int, int]]]]
            - 'style': Optional[str]

    Returns:
        an swu sign string

    Example:
        >>> swu_compose_sign({
        ...     'sequence': ['񀀒','񀀚','񋚥','񋛩'],
        ...     'box': '𝠃',
        ...     'max': [525, 535],
        ...     'spatials': [
        ...         {'symbol': '񋛩', 'coord': [483, 510]},
        ...         {'symbol': '񀀒', 'coord': [501, 466]},
        ...         {'symbol': '񋚥', 'coord': [510, 500]},
        ...         {'symbol': '񀀚', 'coord': [476, 475]}
        ...     ],
        ...     'style': '-C'
        ... })
        '𝠀񀀒񀀚񋚥񋛩𝠃𝤟𝤩񋛩𝣵𝤐񀀒𝤇𝣤񋚥𝤐𝤆񀀚𝣮𝣭-C'
    """
    box = swu_sign_object.get("box", "")
    box_match = re.match(f"^({swu_pattern_box})$", box)
    if not box_match:
        return None
    box = box_match.group(1)

    max_coord = swu_sign_object.get("max")
    max_str = ""
    if max_coord and isinstance(max_coord, list) and len(max_coord) == 2:
        max_str = coord_to_swu(max_coord)
    if not max_str:
        return None

    prefix = ""
    sequence = swu_sign_object.get("sequence")
    if sequence and isinstance(sequence, list):
        prefix_parts = []
        for key in sequence:
            key_match = re.match(f"^({swu_pattern_null_or_symbol})$", key)
            if key_match:
                prefix_parts.append(key_match.group(1))
        prefix = "\U0001D800" + "".join(prefix_parts) if prefix_parts else ""

    signbox = ""
    spatials = swu_sign_object.get("spatials")
    if spatials and isinstance(spatials, list):
        signbox_parts = []
        for spatial in spatials:
            sym = spatial.get("symbol")
            if not isinstance(sym, str):
                continue
            sym_match = re.match(f"^({swu_pattern_symbol})$", sym)
            if not sym_match:
                continue
            sym = sym_match.group(1)

            coord = spatial.get("coord")
            coord_str = ""
            if coord and isinstance(coord, list) and len(coord) == 2:
                coord_str = coord_to_swu(coord)
            if not coord_str:
                continue

            signbox_parts.append(sym + coord_str)
        signbox = "".join(signbox_parts)

    style_str = swu_sign_object.get("style")
    style_candidate = ""
    if isinstance(style_str, str):
        style_match = re.match(f"^({style_pattern_full})$", style_str)
        if style_match:
            style_candidate = style_match.group(1)

    return prefix + box + max_str + signbox + style_candidate


# ----------------------------
# SWU Info
# ----------------------------


def swu_info(swu: str) -> Dict[str, Any]:
    """
    Function to gather sizing information about an swu sign or symbol.

    Args:
        swu: an swu sign or symbol

    Returns:
        information about the swu string

    Example:
        >>> swu_info('𝠀񁲡񈩧𝠂𝤘𝤣񁲡𝣳𝣩񈩧𝤉𝣻-P10Z2')
        {
            'minX': 481,
            'minY': 471,
            'width': 37,
            'height': 58,
            'lane': -1,
            'padding': 10,
            'segment': 'sign',
            'zoom': 2
        }
    """
    lanes = {"𝠁": 0, "𝠂": -1, "𝠃": 0, "𝠄": 1}
    parsed_sign = swu_parse_sign(swu)
    width: Optional[int] = None
    height: Optional[int] = None
    segment: str = "none"
    x1: int = 490
    y1: int = 490
    lane: int = 0
    parsed: Dict[str, Any] = {}  # Unified parsed object for style

    if spatials := parsed_sign.get("spatials"):
        x_coords = [spatial["coord"][0] for spatial in spatials]
        y_coords = [spatial["coord"][1] for spatial in spatials]
        x1 = min(x_coords)
        y1 = min(y_coords)
        x2, y2 = parsed_sign["max"]
        width = x2 - x1
        height = y2 - y1
        segment = "sign"
        lane = lanes[parsed_sign["box"]]
        parsed = parsed_sign
    else:
        parsed_symbol = swu_parse_symbol(swu)
        lane = lanes["𝠃"]
        if coord := parsed_symbol.get("coord"):
            x1, y1 = coord
            width = (500 - x1) * 2
            height = (500 - y1) * 2
            segment = "symbol"
        parsed = parsed_symbol

    style = style_parse(parsed.get("style", ""))
    zoom = style.get("zoom", 1)
    padding = style.get("padding", 0)

    return {
        "minX": x1,
        "minY": y1,
        "width": width or 20,
        "height": height or 20,
        "segment": segment,
        "lane": lane,
        "padding": padding,
        "zoom": zoom,
    }


# ----------------------------
# SWU Columns
# ----------------------------

swu_column_defaults: Dict[str, Any] = {
    "height": 500,
    "width": 150,
    "offset": 50,
    "pad": 20,
    "margin": 5,
    "dynamic": False,
    "background": None,
    "punctuation": {
        "spacing": True,
        "pad": 30,
        "pull": True,
    },
    "style": {
        "detail": ["black", "white"],
        "zoom": 1,
    },
}


def swu_column_defaults_merge(
    options: Optional[Dict[str, Any]] = None
) -> Dict[str, Any]:
    """
    Function to merge an object of column options with default values.

    Args:
        options: object of column options

    Returns:
        object of column options merged with column defaults

    Example:
        >>> swu_column_defaults_merge({'height': 500, 'width': 150})
        {'height': 500, 'width': 150, 'offset': 50, ...}
    """
    if not isinstance(options, dict):
        options = {}
    merged = {**swu_column_defaults, **options}
    merged["punctuation"] = {
        **swu_column_defaults["punctuation"],
        **(options.get("punctuation") or {}),
    }
    merged["style"] = {**swu_column_defaults["style"], **(options.get("style") or {})}
    return drop_none(merged)


def swu_columns(
    swu_text: str, options: Optional[Dict[str, Any]] = None
) -> Dict[str, Any]:
    """
    Function to transform an FSW text to an array of columns.

    Args:
        swu_text: FSW text of signs and punctuation
        options: object of column options

    Returns:
        object of column options, widths array, and column data

    Example:
        >>> swu_columns('𝠀񁲡񈩧𝠃𝤘𝤣񁲡𝣳𝣩񈩧𝤉𝣻 𝠀񃊢񃊫񋛕񆇡𝠃𝤘𝤧񃊫𝣻𝤕񃊢𝣴𝣼񆇡𝤎𝤂񋛕𝤆𝣦 񏌁𝣢𝤂', {'height': 500, 'width': 150})
        {'options': {...}, 'widths': [150], 'columns': [[{'x': 56, 'y': 20, ...}, ...]]}
    """
    if not isinstance(swu_text, str):
        return {}
    values = swu_column_defaults_merge(options)
    input_text = swu_parse_text(swu_text)
    if not input_text:
        return {}

    cursor = 0
    cols: List[List[Dict[str, Any]]] = []
    col: List[Dict[str, Any]] = []
    plus = 0
    center = values["width"] // 2
    max_height = values["height"] - values["margin"]
    pullable = True
    finalize = False

    for val in input_text:
        informed = swu_info(val)
        cursor += plus
        if values["punctuation"]["spacing"]:
            cursor += values["pad"] if informed["segment"] == "sign" else 0
        else:
            cursor += values["pad"]

        finalize = (cursor + informed["height"]) > max_height

        if (
            informed["segment"] == "symbol"
            and values["punctuation"]["pull"]
            and pullable
        ):
            finalize = False
            pullable = False

        if not col:
            finalize = False

        if finalize:
            cursor = values["pad"]
            cols.append(col)
            col = []
            pullable = True

        item = {**informed, "text": val}
        item["x"] = (
            center
            + (values["offset"] * informed["lane"])
            - ((500 - informed["minX"]) * informed["zoom"] * values["style"]["zoom"])
        )
        item["y"] = cursor
        col.append(item)
        cursor += informed["height"] * informed["zoom"] * values["style"]["zoom"]

        if values["punctuation"]["spacing"]:
            plus = (
                values["pad"]
                if informed["segment"] == "sign"
                else values["punctuation"]["pad"]
            )
        else:
            plus = values["pad"]

    if col:
        cols.append(col)

    # Over-height adjustment for pulled punctuation
    if values["punctuation"]["pull"]:
        for c in cols:
            last = c[-1]
            diff = (last["y"] + last["height"]) - (values["height"] - values["margin"])
            if diff > 0:
                adj = (diff // len(c)) + 1
                for i, item in enumerate(c):
                    item["y"] -= adj * i + adj

    # Contract, expand, adjust widths
    widths: List[int] = []
    for c in cols:
        mins = [center - values["offset"] - values["pad"]]
        maxs = [center + values["offset"] + values["pad"]]
        for item in c:
            mins.append(item["x"] - values["pad"])
            maxs.append(item["x"] + item["width"] + values["pad"])
        min_val = min(mins)
        max_val = max(maxs)

        width = values["width"]
        adj = 0
        if not values["dynamic"]:
            adj = center - ((min_val + max_val) // 2)
        else:
            width = max_val - min_val
            adj = -min_val

        for item in c:
            item["x"] += adj
        widths.append(width)

    return {"options": values, "widths": widths, "columns": cols}


__all__ = [
    "swu_structure_kind",
    "swu_structure_category",
    "swu_structure_group",
    "swu_structure_ranges",
    "swu_is_type",
    "swu_colors",
    "swu_colorize",
    "swu_parse_symbol",
    "swu_parse_sign",
    "swu_parse_text",
    "swu_compose_symbol",
    "swu_compose_sign",
    "swu_info",
    "swu_column_defaults",
    "swu_column_defaults_merge",
    "swu_columns",
]
